﻿--[[
Uber credits must be given to nymbia for the metatables and frames, thanks!
]]
local BS = AceLibrary('Babble-Spell-2.2')
local L = AceLibrary('AceLocale-2.2'):new('ClassTimer')

local sm = AceLibrary('SharedMedia-1.0')

local self = ClassTimer
local ClassTimer = ClassTimer

local GetTime = GetTime
local unpack = unpack
local table_sort = table.sort
local math_ceil = math.ceil
local gsub = gsub

local db
ClassTimer.locked = { target = true, player = true, focus = true }

local new, del
do
	local cache = setmetatable({}, {__mode='k'})
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
end
local OnUpdate
do
	local min = L['%dm']
	local function timenum(num)
		if num <= 10 then
			return ('%.1f'):format(num)
		elseif num <= 60 then
			return ('%d'):format(num)
		else
			return min:format(math_ceil(num / 60))
		end
	end
	function OnUpdate(frame)
		local currentTime = GetTime()
		local endTime = frame.endTime
		if currentTime > endTime then
			if not ClassTimer.locked[frame.unit] then ClassTimer.locked[frame.unit] = true end
			self:UpdateUnitBars(frame.unit)
		else
			local remaining = (currentTime - frame.startTime)
			frame:SetValue(endTime - remaining)
			frame.timetext:SetText(timenum(endTime - currentTime))
		end
		
	end
end

local framefactory = {
	__index = function(t,k)
		local bar = CreateFrame('StatusBar', nil, UIParent)
		t[k] = bar
		bar:SetFrameStrata('MEDIUM')
		bar:Hide()
		bar:SetScript('OnUpdate', OnUpdate)
		bar:SetBackdrop({bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background', tile = true, tileSize = 16})
		bar:SetMovable(true)
		bar.text = bar:CreateFontString(nil, 'OVERLAY')
		bar.timetext = bar:CreateFontString(nil, 'OVERLAY')
		bar.icon = bar:CreateTexture(nil, 'DIALOG')
		
		self:ApplySettings()
		return bar
	end
}

ClassTimer.bars = {
	target = setmetatable({}, framefactory),
	focus = setmetatable({}, framefactory),
	player = setmetatable({}, framefactory)
}

ClassTimer.Rules = {}
ClassTimer.RuleMenu = {}
ClassTimer.defaults = {
	Abilities = {},
	Units = {	
		['**'] = {
			enable = true,
			buffs = true,
			debuffs = true,
			growup = false,
			showIcons = false,
			icons = true,
			iconSide = L['Left'],
			scale = 1,

			spacing = 1,
		},
	},
	nametext = true,
	timetext = true,
	
	AllInOne = false,
	texture = 'Blizzard',
	width = 150,
	height = 16,
	font = 'Friz Quadrata TT',
	fontsize = 9,
	alpha = 1,
	scale = 1,
	bartext = '%s (%a) (%u)',
	
	buffcolor = {0,0.49, 1, 1},
	debuffcolor = {1.0,0.7, 0, 1},
	backgroundcolor = {0,0, 0, 1},
		
	textcolor = {1,1,1},
}

function ClassTimer:RulesInit()
local timerargs = {}
local table =  self:CreateTimers()
table['Trinkets'] = self:Trinkets()
table['Race'] = self:Races()
	for k, v in pairs(table) do
		local k = L[k]
		timerargs[k] = {
			type = 'group',
			order = 4,
			name = k,
			desc = L['Which Buffs to show.'],
			args = {}
		}
		for _, m in pairs (v) do
			local m = BS[m]
			ClassTimer.defaults.Abilities[m] = true
			timerargs[k].args[m] = {
				order = 3,
				type = 'toggle',
				name = m,
				desc = L['Enable/Dissable the timer'],
				get = function()
					return self.db.profile.Abilities[m]
				end,
				set = function(u)
					self.db.profile.Abilities[m] = u
				end,
			}
		end
	end
	self.Menu.args.Timers = { type = 'group', order = 1, name = L['Timers'], desc = L['Enable or disable timers'], args = timerargs, disabled = '~IsActive' }
end

function ClassTimer:RulesSetup()
	self:RegisterBucketEvent('UNIT_AURA', 0.5)
	self:RegisterEvent('PLAYER_TARGET_CHANGED')
	self:RegisterEvent('PLAYER_FOCUS_CHANGED')
	self:RegisterEvent('SharedMedia_SetGlobal', function(mtype, override)
		if mtype == 'statusbar' then
			for _, k in pairs(ClassTimer.bars) do
				for _, v in pairs(k) do
					v:SetStatusBarTexture(sm:Fetch('statusbar', override))
				end
			end
		end
	end)
	ClassTimer:ApplySettings()
end

function ClassTimer:OnDisable()
	for _, k in pairs(ClassTimer.bars) do
		for _, v in pairs(k) do
			v:Hide()
		end
	end
end

function ClassTimer:PLAYER_TARGET_CHANGED()
	self:UpdateUnitBars('target', self.db.profile.AllInOne)
end

function ClassTimer:PLAYER_FOCUS_CHANGED()
	self:UpdateUnitBars('focus', self.db.profile.AllInOne)
end
function ClassTimer:UNIT_AURA(units)
	for unit in pairs(units) do
		if self.db.profile.AllInOne then
			self:UpdateUnitBars('player', true)
			return
		end
		self:UpdateUnitBars(unit)
	end
end

do
	local function sort(a,b)
		return a.remaining < b.remaining
	end
	local function text(spell, apps, unit)
		local str = self.db.profile.bartext
		str = gsub(str, '%%s', spell)
		str = gsub(str, '%%u', unit)
		
		if apps and apps > 1 then
			str = gsub(str, '%%a', apps)
		else
			str = gsub(str, '%%a', '')
		end
		str = gsub(str, '%%n', GetUnitName(unit))
		str = gsub(str, '%s%s+', ' ')
		str = gsub(str, '%p%p+', '')	
		return str
	end
	local tmp = {}
	local called = false -- prevent recursive calls when new ClassTimer.bars are created.
	function ClassTimer:GetBuffs(unit)
		if self.db.profile.Units[unit].buffs then
				for i = 1, 32 do
					local name, _, texture, count, duration, remaining = UnitBuff(unit, i)
					if not name then
						break
					end
					if duration and duration > 0 and self.db.profile.Abilities[name] then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.unit = unit
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.count = count
						t.isbuff = true
					end
				end
			end
			if self.db.profile.Units[unit].debuffs then
				for i = 1, 40 do
					local name, _, texture, count, _, duration, remaining = UnitDebuff(unit, i)
					if not name then
						break
					end
					if duration and duration > 0 and self.db.profile.Abilities[name] then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.unit = unit
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.count = count
					end
				end
			end
	end
	function ClassTimer:UpdateUnitBars(unit, all)
	if not ClassTimer.locked[unit] then return end
	if all and unit ~= 'player' then unit = 'player' end
	--[[for _, k in pairs(unitTable) do
		if k ~= unit and unit ~= 'target' and UnitIsUnit(unit, k) then
			return
		end
	end]]
	if not ClassTimer.bars[unit] then return end
		if called then
			return
		end
		called = true
		local db = self.db.profile
		if db.Units[unit].enable then
			local currentTime = GetTime()
			for k in pairs(tmp) do
				tmp[k] = del(tmp[k])
			end
			self:GetBuffs(unit)
			if all then
				if not UnitIsUnit('target', unit) then self:GetBuffs('target') end
				if not UnitIsUnit('focus', unit) then self:GetBuffs('focus') end
			end
			table_sort(tmp, sort)
			local maxindex = 0
			for k,v in ipairs(tmp) do
				maxindex = k
				local bar = ClassTimer.bars[unit][k]
				
				bar.text:SetText(text(v.name, v.count, v.unit))
				bar.icon:SetTexture(v.texture)
				local elapsed = (v.duration - v.remaining)
				local startTime, endTime = (currentTime - elapsed), (currentTime + v.remaining)
				bar.startTime = startTime
				bar.unit = unit
				bar.endTime = endTime
				bar:SetMinMaxValues(startTime, endTime)
				bar:Show()
				bar.isbuff = isbuff
			end
			for i = maxindex+1, #ClassTimer.bars[unit] do
				ClassTimer.bars[unit][i]:Hide()
			end
		else
			for _, v in ipairs(ClassTimer.bars[unit]) do
				v:Hide()
			end
		end
		called = false
	end
end
do
	local function apply(unit, i, bar, db)
		local position, icons, ShowIcons, iconside, gap, spacing, direction, width
		local bars = ClassTimer.bars[unit]
		growup = db.Units[unit].growup
		icons = db.Units[unit].icons
		showIcons = db.Units[unit].showIcons 
		iconSide = db.Units[unit].iconSide
		gap = db.Units[unit].gap
		spacing = db.Units[unit].spacing
		offset = db.Units[unit].offset
		bar:ClearAllPoints()
		bar:SetStatusBarTexture(sm:Fetch('statusbar', db.texture))
		if showIcons then
			bar:SetStatusBarColor(1, 1, 1, 0)
			bar:SetWidth(1)
		else 
			bar:SetWidth(db.width)
		end
		bar:SetHeight(db.height)
		bar:SetScale(db.Units[unit].scale)
		bar:SetAlpha(db.alpha)
		
		if bar.isbuff then
			bar:SetStatusBarColor(unpack(db.buffcolor))
		else
			bar:SetStatusBarColor(unpack(db.debuffcolor))
		end
			bar:SetBackdropColor(unpack(db.backgroundcolor))
		if showIcons then
			bar.text:SetFont(sm:Fetch('font', db.font), db.fontsize)
		end
		if i == 1 then
				if db.Units[unit].x then
					bar:SetPoint('BOTTOMLEFT', UIParent, 'BOTTOMLEFT', db.Units[unit].x, db.Units[unit].y)
				else
					bar:SetPoint('CENTER', UIParent)
				end
		else
			if growup then
				bar:SetPoint('BOTTOM', bars[i-1], 'TOP', 0, spacing)
			else 
				bar:SetPoint('TOP', bars[i-1], 'BOTTOM', 0, -1 * spacing)
			end
		end
		
		local timetext = bar.timetext
		if db.timetext then
			timetext:Show()
			timetext:ClearAllPoints()
			timetext:SetFont(sm:Fetch('font', db.font), db.fontsize)
			timetext:SetShadowColor( 0, 0, 0, 1)
			timetext:SetShadowOffset( 0.8, -0.8 )
			timetext:SetTextColor(unpack(db.textcolor))
			timetext:SetNonSpaceWrap(false)
			timetext:ClearAllPoints()
			timetext:SetWidth(db.width)
			timetext:SetPoint('LEFT', bar, 'LEFT', -2, 0)
			timetext:SetJustifyH('RIGHT')
		else
			timetext:Hide()
		end
		
		local text = bar.text
		if db.nametext then
			text:Show()
			text:ClearAllPoints()
			text:SetFont(sm:Fetch('font', db.font), db.fontsize)
			text:SetShadowColor( 0, 0, 0, 1)
			text:SetShadowOffset( 0.8, -0.8 )
			text:SetTextColor(unpack(db.textcolor))
			text:SetNonSpaceWrap(false)
			text:ClearAllPoints()
			text:SetWidth(db.width - db.fontsize * 2.5)
			text:SetHeight(db.height)
			text:SetPoint('LEFT', bar, 'LEFT', 2, 0)
			text:SetJustifyH('LEFT')
		else
			text:Hide()
		end

		local icon = bar.icon
		if icons then
			icon:Show()
			icon:SetWidth(db.height-1)
			icon:SetHeight(db.height-1)
			icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
			icon:ClearAllPoints()
			if iconSide == L['Left'] then
				icon:SetPoint('RIGHT', bar, 'LEFT', -1, 0)
			else
				icon:SetPoint('LEFT', bar, 'RIGHT', 1, 0)
			end
		else
			icon:Hide()
		end
		
		return direction
	end
	function ClassTimer:ApplySettings()
		for n, k in pairs(ClassTimer.bars) do
				for u, v in pairs(k) do
					apply(n, u, v, ClassTimer.db.profile)
				end
			end
	end
end
